/**
 * @file cuesheet.h
 * @brief פʣ
 * @author BananaJinn
 * @version $Id: cuesheet.c,v 1.2 2010/01/19 14:54:19 bananajinn Exp $
 * ʣ̲
 * Copyright (C) 2004-2010 BananaJinn<banana@mxh.mesh.ne.jp>.
 */
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "mem.h"
#include "aspi.h"
#include "struct.h"
#include "cmd.h"
#include "log.h"
#include "discinfo.h"

#define SEND_GAPDATA 1

/**
 * Next Writable Address 
 * @param[in] drive ֹ¤
 * @param[in] track_num ȥåֹ
 * @param[out] nwa_ret NWA
 * @retval RET_OK ｪλ
 * @retval RET_NG 顼
 */
static int GetNWA(CMDDRIVE *drive, WORD track_num, DWORD *nwa_ret)
{
	DWORD nwa = -150;
	int ret;
	struct _TRACKINFO *ti;
	
	if(REALDRIVE(drive)){
		ret = SendReadTrackInfo(drive, track_num);
		if(ret!=RET_OK){
			DispCommandError(drive);
			return ret;
		}
		ti = (struct _TRACKINFO *)drive->data_buf;
		nwa = Get4bytes(ti->next_writable_addr);
		if(nwa == (DWORD)0){
			nwa -= 150;
		}
	}
	*nwa_ret = nwa;
	return RET_OK;
}

/**
 * 塼ѤΥ
 * @param[in,out] cs 塼
 * @param[in,out] num_cs 塼ȤΥȥ
 * @param[in] num ݤ륨ȥ
 * @return ݤΰΥݥ󥿤֤顼ξ NULL ֤
 */
static struct _CUESHEET *NewSheet(struct _CUESHEET **cs, int *num_cs, int num)
{
	struct _CUESHEET *result = NULL;
	*cs = (struct _CUESHEET *)MemResize(
		*cs, (*num_cs+num)*sizeof(struct _CUESHEET));
	if(*cs != NULL){
		result = &(*cs)[*num_cs];
		*num_cs += num;
	}
	return result;
}

/**
 * ǥֹ
 * @param[in,out] cs 塼
 * @param[in,out] num_cs 塼ȤΥȥ
 * @param[in] discinfo ǥ
 * @retval RET_OK	ｪλ
 * @retval RET_MEMERR ­
 */
static int SetMCN(struct _CUESHEET **cs, int *num_cs, CPDISCINFO *discinfo)
{
	DebugLog("CreateCueSheet: Media catalog number.\n");
	struct _CUESHEET *csp = NewSheet(cs, num_cs, 2);
	if(csp == NULL){
		return RET_MEMERR;
	}

	csp->ctladr = CSADR_MCN;
	memcpy(&csp->tno, discinfo->media_catalog_number, 7);
	csp++;
	csp->ctladr = CSADR_MCN;
	memcpy(&csp->tno, discinfo->media_catalog_number+7, 7);
	return RET_OK;
}


/**
 * ISRC
 * @param[in,out] cs 塼
 * @param[in,out] num_cs 塼ȤΥȥ
 * @param[in] track_num ȥåֹ
 * @param[in] cpti ȥå
 * @retval RET_OK	ｪλ
 * @retval RET_MEMERR ­
 */
static int SetISRC(struct _CUESHEET **cs, int *num_cs, WORD track_num,
				   CPTRACKINFO *cpti)
{
	DebugLog("CreateCueSheet: ISRC.\n");
	struct _CUESHEET *csp = NewSheet(cs, num_cs, 2);
	if(csp == NULL){
		return RET_MEMERR;
	}

	csp->ctladr = CSADR_ISRC;
	csp->tno = (BYTE)track_num;
	memcpy(&csp->index, cpti->isrc, 6);
	csp++;
	csp->ctladr = CSADR_ISRC;
	csp->tno = (BYTE)track_num;
	memcpy(&csp->index, cpti->isrc+6, 6);
	return RET_OK;
}


/**
 * Lead-in
 * @param[in,out] cs 塼
 * @param[in,out] num_cs 塼ȤΥȥ
 * @param[in] cpti ȥå
 * @param[in] cdtext CD-TEXTͭե饰
 * @param[in] nwa NWA
 * @retval RET_OK	ｪλ
 * @retval RET_MEMERR ­
 */
static int SetLeadIn(struct _CUESHEET **cs, int *num_cs, CPTRACKINFO *cpti,
					 BOOL cdtext, DWORD nwa)
{
	DebugLog("CreateCueSheet: Lead-in.\n");
	struct _CUESHEET *csp = NewSheet(cs, num_cs, 1);
	if(csp == NULL){
		return RET_MEMERR;
	}
	csp->ctladr = IS_TRACKMODE_DATA(cpti->trackinfo.track_mode)
		? CSCTL_DATA : CSCTL_AUDIO;
	csp->ctladr |= CSADR_NORMAL;
	csp->tno = 0;
	csp->index = 0;
	csp->dataform = cdtext ? 0x41 : 0x01;
	csp->scms = 0;
	/*LBA2MSF(nwa, &csp->min, &csp->sec, &csp->frame);*/
	/** Lead-in Ͼ0 */
	csp->min = 0;
	csp->sec = 0;
	csp->frame = 0;
	return RET_OK;
}

/**
 * Lead-out
 * @param[in,out] cs 塼
 * @param[in,out] num_cs 塼ȤΥȥ
 * @param[in] cpti ȥå
 * @param[in] nwa NWA
 * @retval RET_OK	ｪλ
 * @retval RET_MEMERR ­
 */
static int SetLeadOut(struct _CUESHEET **cs, int *num_cs,
					  CPTRACKINFO *cpti, DWORD nwa)
{
	DebugLog("CreateCueSheet: Lead-out.\n");
	struct _CUESHEET *csp = NewSheet(cs, num_cs, 1);
	if(csp == NULL){
		return RET_MEMERR;
	}
	csp->ctladr = IS_TRACKMODE_DATA(cpti->trackinfo.track_mode)
		? CSCTL_DATA : CSCTL_AUDIO;
	csp->ctladr |= CSADR_NORMAL;
	csp->tno = 0xAA;
	csp->index = 1;
	if(!IS_TRACKMODE_DATA(cpti->trackinfo.track_mode))
		csp->dataform = 0x01;
	else
		csp->dataform = cpti->mode2 ? 0x24 : 0x14;
	csp->scms = cpti->trackinfo.copy ? 0x80 : 0;
	LBA2MSF(nwa, &csp->min, &csp->sec, &csp->frame);
	return RET_OK;
}


/**
 * ץꥮå
 * @param[in,out] cs 塼
 * @param[in,out] num_cs 塼ȤΥȥ
 * @param[in] track_num ȥåֹ
 * @param[in] lcpti ľȥå
 * @param[in] cpti ߥȥå
 * @param[in] gap_len ץꥮåĹ
 * @param[in] twopart 2ѡȥץꥮåפɤ
 * @param[in,out] nwa NWA
 * @retval RET_OK	ｪλ
 * @retval RET_MEMERR ­
 */
static int SetPreGap(struct _CUESHEET **cs, int *num_cs, WORD track_num,
					 CPTRACKINFO *lcpti, CPTRACKINFO *cpti,
					 DWORD gap_len, BOOL twopart, DWORD *nwa)
{
	struct _CUESHEET *csp = NULL;
	DebugLog("CreateCueSheet: PreGap. track=%d gap=%ld nwa=0x%08lX\n",
			 track_num, (long)gap_len, *nwa);
	if(gap_len > 0){
		if(twopart && gap_len > 150){
			csp = NewSheet(cs, num_cs, 1);
			if(csp == NULL){
				return RET_MEMERR;
			}
			csp->ctladr =
				IS_TRACKMODE_DATA(lcpti->trackinfo.track_mode)
				? CSCTL_DATA : CSCTL_AUDIO;
			csp->ctladr |= CSADR_NORMAL;
			csp->tno = (BYTE)track_num;
			csp->index = 0;
#if SEND_GAPDATA
			if(!IS_TRACKMODE_DATA(lcpti->trackinfo.track_mode)){
				csp->dataform = 0x00;
			}
			else{
				csp->dataform = lcpti->mode2 ? 0x20 : 0x10;
			}
#else
			if(!IS_TRACKMODE_DATA(lcpti->trackinfo.track_mode)){
				csp->dataform = 0x01;
			}
			else{
				csp->dataform = lcpti->mode2 ? 0x24 : 0x14;
			}
#endif
			csp->scms = lcpti->trackinfo.copy ? 0x80 : 0;
			LBA2MSF(*nwa, &csp->min, &csp->sec, &csp->frame);
			*nwa += (gap_len-150);
			gap_len = 150;
		}
		csp = NewSheet(cs, num_cs, 1);
		if(csp == NULL){
			return RET_MEMERR;
		}
		csp->ctladr = IS_TRACKMODE_DATA(cpti->trackinfo.track_mode)
			? CSCTL_DATA : CSCTL_AUDIO;
		csp->ctladr |= CSADR_NORMAL;
		csp->tno = (BYTE)track_num;
		csp->index = 0;
#if SEND_GAPDATA
		if(!IS_TRACKMODE_DATA(cpti->trackinfo.track_mode)){
			csp->dataform = 0x00;
		}
		else{
			csp->dataform = cpti->mode2 ? 0x20 : 0x10;
		}
#else
		if(!IS_TRACKMODE_DATA(cpti->trackinfo.track_mode)){
			csp->dataform = 0x01;
		}
		else{
			csp->dataform = cpti->mode2 ? 0x24 : 0x14;
		}
#endif
		csp->scms = cpti->trackinfo.copy ? 0x80 : 0;
		LBA2MSF(*nwa, &csp->min, &csp->sec, &csp->frame);
		*nwa += gap_len;
	}
	return RET_OK;
}


/**
 * Audioޤϥǡȥå
 * @param[in,out] cs 塼
 * @param[in,out] num_cs 塼ȤΥȥ
 * @param[in] track_num ȥåֹ
 * @param[in] cpti ȥå
 * @param[in,out] nwa NWA
 * @retval RET_OK	ｪλ
 * @retval RET_MEMERR ­
 */
static int SetTrack(struct _CUESHEET **cs, int *num_cs,
					WORD track_num, CPTRACKINFO *cpti, DWORD *nwa)
{
	DebugLog("CreateCueSheet: Track%d.\n", track_num);
	struct _CUESHEET *csp = NewSheet(cs, num_cs, 1);
	if(csp == NULL){
		return RET_MEMERR;
	}
	csp->ctladr = IS_TRACKMODE_DATA(cpti->trackinfo.track_mode)
		? CSCTL_DATA : CSCTL_AUDIO;
	csp->ctladr |= CSADR_NORMAL;
	csp->tno = (BYTE)track_num;
	csp->index = 1;
	if(!IS_TRACKMODE_DATA(cpti->trackinfo.track_mode))
		csp->dataform = 0x00;
	else
		csp->dataform = cpti->mode2 ? 0x20 : 0x10;
	csp->scms = cpti->trackinfo.copy ? 0x80 : 0;
	LBA2MSF(*nwa, &csp->min, &csp->sec, &csp->frame);
	*nwa += Get4bytes(cpti->trackinfo.track_size);

	return RET_OK;
}

/**
 * SAOϿCueSheetκ
 * @param[in]	drive	ֹ¤
 * @param[in]	discinfo  ǥ
 * @param[in]	track_start ϥȥåֹ
 * @param[out]	cs_ret	CueSheet
 * @param[out]  numcs_ret CueSheetǿ
 * @param[in]   enable_cdtext ⤷CD-TEXTͭˤ뤫ɤ
 * @retval	RET_OK	ｪλ
 * @retval	RET_NG	顼
 */
int CreateCueSheet(CMDDRIVE *drive, CPDISCINFO *discinfo,
				   WORD track_start,
				   struct _CUESHEET **cs_ret, int *numcs_ret,
				   BOOL enable_cdtext)
{
	struct _CUESHEET *cs = NULL;
	int cur_cs = 0;
	int num_cs = 0;
	DWORD gap_len;
	WORD sess_num = 0;
	DWORD nwa;
	CPTRACKINFO *cpti=NULL, *lcpti=NULL;
	int ret;
	BOOL gap_2part;
	BOOL cdtext=FALSE;
	WORD track_num;

	if(track_start==1 && discinfo->cdtext_size>0 && enable_cdtext){
		cdtext = TRUE;
	}

	track_num = track_start;
	cur_cs = 0;
	num_cs = 0;
	sess_num = 0;
	
	ret = GetNWA(drive, track_num, &nwa);
	if(ret != RET_OK){
		return ret;
	}

	if(strlen((char *)discinfo->media_catalog_number)!=0){
		/* Media Catalog Number */
		ret = SetMCN(&cs, &num_cs, discinfo);
		if(ret != RET_OK){
			return ret;
		}
	}
	
	cpti = NULL;
	lcpti = NULL;
	ret = RET_OK;
	for( ; track_num <= discinfo->tracks; track_num++){
		lcpti = cpti;
		cpti = &discinfo->trackinfo[track_num-1];
		gap_len = 0;
		if(sess_num==0){
			sess_num = SESSION_NUMBER(&cpti->trackinfo);
			ret = SetLeadIn(&cs, &num_cs, cpti, cdtext, nwa);
			if(ret != RET_OK){
				return ret;
			}
			if(sess_num >= 2){
				nwa -= 150;	/* pregapʬ */
			}
			gap_len = Get4bytes(cpti->trackinfo.track_start) - nwa;
		}
		else if(sess_num != SESSION_NUMBER(&cpti->trackinfo)){
			cpti = lcpti;
			break;
		}

		if(strlen((char *)cpti->isrc)!=0){
			ret = SetISRC(&cs, &num_cs, track_num, cpti);
			if(ret != RET_OK){
				return ret;
			}
		}
		gap_2part = FALSE;
		if(gap_len==0){
#if 1	/* äɤ? */
			if(lcpti!=NULL){
				gap_len = Get4bytes(cpti->trackinfo.track_start) -
					(Get4bytes(lcpti->trackinfo.track_start) +
					 Get4bytes(lcpti->trackinfo.track_size));
				if(IS_TRACKMODE_DATA(cpti->trackinfo.track_mode)){
					/* DATA */
					if(!IS_TRACKMODE_DATA(lcpti->trackinfo.track_mode) ||
					   lcpti->mode2 != cpti->mode2){
						/* 2part */
						DebugLog("CreateCueSheet: Two part pregap.(cdda -> data)\n");
						gap_2part = TRUE;
					}
				}
				else{
					/* AUDIO */
					if(IS_TRACKMODE_DATA(lcpti->trackinfo.track_mode)){
						/* DATAξ⤫? */
						DebugLog("CreateCueSheet: Two part pregap.(data -> cdda)\n");
						gap_2part = TRUE;
					}
				}
			}
#else
			if(!IS_TRACKMODE_DATA(cpti->trackinfo.track_mode)){
				/* AUDIO */
				gap_len = cpti->pause_len;
			}
			else if(lcpti!=NULL){
				/* DATA  2nd ʹ */
				gap_len = 150;
				if(!IS_TRACKMODE_DATA(lcpti->trackinfo.track_mode) ||
				   lcpti->mode2 != cpti->mode2){
					/* 2part */
					gap_len = 75+150;
					gap_2part = TRUE;
				}
			}
#endif
		}
		nwa = Get4bytes(cpti->trackinfo.track_start) - gap_len;
		ret = SetPreGap(&cs, &num_cs, track_num, lcpti, cpti,
						gap_len, gap_2part, &nwa);
		if(ret != RET_OK){
			return ret;
		}
		/* Audio/Data */
		ret = SetTrack(&cs, &num_cs, track_num, cpti, &nwa);
		if(ret != RET_OK){
			return ret;
		}
	}
	/* Lead-out */
	ret = SetLeadOut(&cs, &num_cs, cpti, nwa);
	if(ret != RET_OK){
		return ret;
	}
	
	*cs_ret = cs;
	*numcs_ret = num_cs;
	DebugLog("CreateCueSheet: Done.\n");
	return RET_OK;
}

